'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.equal = equal;
exports.allEqual = allEqual;
exports.deepEqual = deepEqual;
exports.includes = includes;
exports.inBounds = inBounds;
exports.callback = callback;
exports.arrayLength = arrayLength;
exports.exists = exists;
exports.bitmask = bitmask;

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * assert strict equal single value
 * @param {string} name   key of the value to test
 * @param {mixed}  value  comparision
 * @throws TypeError assertion-error
 */
function equal(name, value) {
    if (this.vars[name] !== value) {
        throw new TypeError('Expected ' + value + ', found ' + this.vars[name] + ' at ' + name);
    }
}

/**
 * assert deep equality each value in Object|Array
 * @param {string} name     key of the object|array to test
 * @param {mixed} testValue comparision, undefined for auto-detect
 * @throws TypeError assertion-error
 */
/**
 * Asserts help you to make sure, the buffer you're parsing
 * is in the correct format. These assertions are like chai, throwing an error
 * when an assertion doesn't hold.
 *
 * These functions won't modify {@link Corrode#vars}.
 */

function allEqual(name, testValue) {
    var values = this.vars[name];
    if (!Array.isArray(values)) {
        values = _lodash2.default.values(values);
    }

    if (typeof testValue === 'undefined') {
        testValue = values[0];
    }

    var notEqualObjects = values.filter(function (varValue) {
        return varValue !== testValue;
    });
    if (notEqualObjects.length !== 0) {
        throw new TypeError('Expected values in ' + JSON.stringify(this.vars[name]) + ' to all be ' + testValue);
    }
}

/**
 * assert equality objects
 * @param {string} name  key of the object to test
 * @param {object} value comparision
 * @throws TypeError assertion-error
 */
function deepEqual(name, value) {
    var binaryValue = this.vars[name];
    if (!_lodash2.default.isEqual(binaryValue, value)) {
        throw new TypeError('Expected ' + JSON.stringify(value) + ', found ' + JSON.stringify(binaryValue));
    }
}

/**
 * assert array|object to contain item
 * @param {string} name key of the value to test
 * @param {array|object} arr comparision
 * @throws TypeError assertion-error
 */
function includes(name, arr) {
    if (!_lodash2.default.includes(arr, this.vars[name])) {
        throw new TypeError('Expected ' + JSON.stringify(arr) + ' to include ' + this.vars[name]);
    }
}

/**
 * assert value to be within the bounds of an array
 * @param {string} name  key of the number to test
 * @param {array} value comparision
 * @throws TypeError assertion-error
 */
function inBounds(name, value) {
    var index = this.vars[name];

    if (index < 0 || index >= value.length) {
        throw new TypeError('Expected Array of ' + value.length + ' items to be at least ' + index + ' long');
    }
}

/**
 * assert value via callback
 * @param {string}   name   key of the value to test
 * @param {function} fn     callback
 * @param {string}   testname optional test-name
 * @throws TypeError assertion-error
 */
function callback(name, fn) {
    var testname = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : fn.name;

    if (!fn(this.vars[name])) {
        throw new TypeError('Callback failed at ' + testname + '(' + this.vars[name] + ')');
    }
}

/**
 * assert array to be a given length
 * @param {string}        name   key of the value to test
 * @param {number|string} length comparision
 * @throws TypeError assertion-error
 */
function arrayLength(name, length) {
    // try to get the length param from the vars if available
    if (typeof length === 'string' && this.vars[length] !== 'undefined') {
        length = this.vars[length];
    }

    if (typeof this.vars[name] === 'undefined' || this.vars[name].length !== length) {
        throw new TypeError('Expected array to have a length of ' + length + ', has ' + this.vars[name].length);
    }
}

/**
 * asserts a variable exists in the first place
 * @param {string}   name   key of the value to test
 * @throws TypeError assertion-error
 */
function exists(name) {
    if (typeof this.vars[name] === 'undefined') {
        throw new TypeError('Expected var ' + name + ' to exist');
    }
}

/**
 * asserts a variable matches a given bitmask
 * @param  {string} name         key of the value to test
 * @param  {number} mask         bitmask to match
 * @param  {boolean} assertMatch true: should match; false: shouldn't match
 * @throws TypeError assertion-error
 */
function bitmask(name, mask) {
    var assertMatch = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : true;

    var val = this.vars[name];

    if ((val & mask) === mask === !assertMatch) {
        throw new TypeError('Expected var ' + name + ' to ' + (assertMatch ? '' : 'not ') + 'match bitmask (value: 0b' + val.toString(2) + ' assert: 0b' + mask.toString(2) + ')');
    }
}
//# sourceMappingURL=index.js.map